import * as Mqtt from 'mqtt'
import * as log from './log'

const MQTT_BROKER_DEFAULT_URL : string = "mqtt://localhost"

type TopicHandlerCallbackFunc = (message: any) => any

class TopicHandlerCallback {
    private _label: string
    private _func: TopicHandlerCallbackFunc
    constructor(label: string, func: TopicHandlerCallbackFunc) {
        this._label = label
        this._func = func
    }
    get func() : TopicHandlerCallbackFunc {
        return this._func
    }
    public toString() : string {
        let funcName : string = (this._func.name === "") ? "lambda" : this._func.name
        return `<${funcName}, ${this._label}>`
    }
}

interface TopicHandler {
    topic: string,
    callbacks: TopicHandlerCallback[]
}

export function passThrough(message: any) {
    return message
}

export class MqttClient {
    private _mqttClient: Mqtt.Client
    private _mqttBrokerUrl: string
    private _topicHandlers: TopicHandler[]

    constructor(mqttBrokerUrl? : string) {
        this._mqttBrokerUrl = (mqttBrokerUrl) ? mqttBrokerUrl : MQTT_BROKER_DEFAULT_URL
        this._topicHandlers = []
    }

    register(topic: string, label: string, callbackFunc: TopicHandlerCallbackFunc) : void {
        let done: boolean = false
        let callback : TopicHandlerCallback = new TopicHandlerCallback(label, callbackFunc)
        for (let topicHandler of this._topicHandlers) {
            if (topicHandler.topic === topic) {
                topicHandler.callbacks.push(callback)
                done = true
                log.info(`additional callback ${callback.toString()} added for topic ${topic}`)
            }
        }
        if (! done) {
            this._topicHandlers.push({topic: topic, callbacks:[ callback ]})
            log.info(`first callback ${callback.toString()} added for topic ${topic}`)
        }
    }

    exec() : void {
        this._mqttClient = Mqtt.connect(this._mqttBrokerUrl)
        this._mqttClient.on('error', log.error)
        this._mqttClient.on('connect', (): void => {
            log.info("connected to mqtt broker")
            for (let topicHandler of this._topicHandlers) {
                this._mqttClient.subscribe(topicHandler.topic)
            }
        })
        this._mqttClient.on('message', (topic: string, payload: Buffer): void => {
            log.info(`message received, topic ${topic}, payload ${payload}`)
            for (let topicHandler of this._topicHandlers) {
                if (this.topicMatch(topicHandler.topic, topic)) {
                    log.info(`received topic ${topic} matches registered topic ${topicHandler.topic}`)
                    // topicHandler.callback(payload)
                    let returnValue: any = payload
                    for (let topicHandlerCallback of topicHandler.callbacks) {
                        log.info(`Calling ${topicHandlerCallback}`)
                        returnValue = topicHandlerCallback.func(returnValue)
                    }
                    log.info(`Final return value is ${returnValue}`)
                }
            }
        })
    }

    private topicMatch(registeredTopic: string, receivedTopic: string) : boolean {
        let registeredTopicFields = registeredTopic.split('/')
        let receivedTopicFields = receivedTopic.split('/')
        for (let field in registeredTopicFields) {
            let regField = registeredTopicFields[field]
            let recvField = receivedTopicFields[field]
            log.info(`recv: ${recvField}, reg: ${regField}`)
            if (regField === "#") {
                log.info('true')
                return true
            }
            if (regField != recvField && regField != "+") {
                log.info('false')
                return false
            }
        }
        if (registeredTopicFields.length == receivedTopicFields.length) {
            log.info('true')
            return true
        } else {
            log.info('false')
            return false
        }

    }
}
